(* Converting syntax trees generated by SableCC into
   manually designed abstract syntax trees in Sabsyn *)

open Speclanguage
open Nodes
open Types

let internal msg = failwith msg

let unsome u = match u with
  Some x -> x
| _ -> internal "Unexpected None option in syntax tree"

let s_extract x = Tokens.getText (unsome x)

let parseIntLit lit = match lit with
| ADecIntLit l1 -> 
    int_of_string (s_extract l1.dec_int_lit_decimal_int_lit)
| _ -> internal "Hexadecimal or octal not handled yet"

(* Value type stuff *)
let rec value_type_from_ugly x =
  match x with
  | AIntType _ -> TInt
  | AFloatType _ -> TFloat
  | ABoolType _ -> TBool
  | ACharType _ -> TChar
  | AStringType _ -> TString
  | AByteType _ -> TByte
  | AFormatType f -> TObj (Tokens.getText (unsome f.format_type_identifier))
  | AArrayType a -> 
      let rec stackers base_type a' = 
        match a' with
        | AOneDims _ -> base_type
        | AManyDims a'' -> stackers (TArray base_type) (unsome a''.many_dims_dims)
      in (TArray (stackers (value_type_from_ugly (unsome a.array_type_type)) (unsome a.array_type_dims)))
  | ATupleType t -> TTuple (List.map value_type_from_ugly t.tuple_type_type)
  | ASetType t -> TSet (value_type_from_ugly (unsome t.set_type_type))

let format_name_from_ugly x =
  match x with
  | AFormatType f -> Id.fetch_format 
        (Tokens.getText (unsome f.format_type_identifier))
  | _ -> failwith "a primitive type"

(******************************)
(* qualified names desugaring *)
(******************************)

let (currentPreds : Sabsyn.predDef list ref) = ref []
let (currentSets : (Id.var_t * Types.value_type) list ref) = ref []
let (currentProps : Id.var_t list ref) = ref []

let qualToId (modName:string) (idName:string) : string =
  modName ^ "." ^ idName

let de_name module_name qualifiable_names (n:p_name) : Sabsyn.ident =
  match n with
  | ASimpleName n1 -> 
      let n1_id = s_extract n1.simple_name_identifier in
      if (List.mem n1_id qualifiable_names) then
        qualToId module_name n1_id
      else n1_id
  | AQualifiedName n1 ->
      qualToId 
        (s_extract n1.qualified_name_module)
        (s_extract n1.qualified_name_set)
        
(*******************************)
(* translating set expressions *)
(*******************************)
     
let rec de_setExp mn qn (s : p_form) : Sabsyn.setExp = 
  match s with
  | AVarForm s1 -> 
      Sabsyn.Prevar (de_name mn qn (unsome s1.var_form_name))
  | APrimedForm s1 ->
      let v = de_setExp mn qn (unsome s1.primed_form_form) in
      (match v with
      | Sabsyn.Prevar i -> Sabsyn.Postvar i
      | _ -> failwith "priming currently allowed only for variables")
  | AEmptyForm _ -> Sabsyn.Emptyset
  | AUnionForm ss ->
      Sabsyn.Union (List.map (de_setExp mn qn) ss.union_form_form)
  | AInterForm ss ->
      Sabsyn.Inter (List.map (de_setExp mn qn) ss.inter_form_form)
  | ADiffForm f1 ->
      Sabsyn.Diff (de_setExp mn qn (unsome f1.diff_form_l),
                   de_setExp mn qn (unsome f1.diff_form_r))
  | _ -> failwith "not a setexp"

(*******************************)
(* translating atomic formulas *)
(*******************************)

let rec de_atom_form mn qn (f : p_form) : Sabsyn.atomForm =
  match f with
  | AEqForm f1 ->
      Sabsyn.Eq (de_setExp mn qn (unsome f1.eq_form_lhs),
             de_setExp mn qn (unsome f1.eq_form_rhs))
  | ANeqForm f1 ->
      Sabsyn.Neq (de_setExp mn qn (unsome f1.neq_form_lhs),
              de_setExp mn qn (unsome f1.neq_form_rhs))
  | ASubForm f1 ->
      Sabsyn.Sub (de_setExp mn qn (unsome f1.sub_form_lhs),
              de_setExp mn qn (unsome f1.sub_form_rhs))
  | ACardeqForm f1 ->
      (let s = unsome f1.cardeq_form_form in
      let i = parseIntLit (unsome f1.cardeq_form_int_lit) in
      Sabsyn.Cardeq (de_setExp mn qn s, i))
  | ACardleqForm f1 ->       
      (let s = unsome f1.cardleq_form_form in
      let i = parseIntLit (unsome f1.cardleq_form_int_lit) in
      Sabsyn.Cardleq (de_setExp mn qn s, i))
  | ACardgeqForm f1 ->
      (let s = unsome f1.cardgeq_form_form in
      let i = parseIntLit (unsome f1.cardgeq_form_int_lit) in
      Sabsyn.Cardgeq (de_setExp mn qn s, i))
  | ADisjointForm f1 ->
      Sabsyn.Disjoint 
        (List.map (de_setExp mn qn) f1.disjoint_form_form)
  | ATrueForm _ -> Sabsyn.True
  | AFalseForm _ -> Sabsyn.False
  | AVarForm s1 -> 
      Sabsyn.Propvar (de_name mn qn (unsome s1.var_form_name))
  | APrimedForm s1 ->
      let v = de_atom_form mn qn (unsome s1.primed_form_form) in
      (match v with
      | Sabsyn.Propvar i -> Sabsyn.Propvarpost i
      | _ -> failwith "priming prop variable?")
  | _ -> failwith "atom: should have gone away"

(************************)
(* translating formulas *)
(************************)

let rec convert_form 
    (mn:string) (s:Id.var_t list) (b:Id.var_t list)
    (f: p_form) : Sabsyn.form =
  let rec convert_form0 (f:p_form) : Sabsyn.form = 
    match f with
    | APredForm f2 ->
        let pn = s_extract f2.pred_form_predicate_identifier in
        let find_pn x = (x.Sabsyn.predDefName = pn) in
        if not (List.exists find_pn !currentPreds) then
          begin
            Util.err "spec" ("undefined pred "^pn);
            Sabsyn.Atom Sabsyn.False
          end
        else
          let f = List.find find_pn !currentPreds in
          let bb = f.Sabsyn.predDefBody in
          let actuals = List.map (de_setExp mn (s @ b)) f2.pred_form_form in
          let fasubst = List.map2 (fun x y -> (x, y))
              f.Sabsyn.predDefFormals actuals in
          List.fold_right Sabsyn.subst fasubst bb
    | APredprimeForm f2 ->
        let pn = s_extract f2.predprime_form_predicate_identifier in
        let find_pn x = (x.Sabsyn.predDefName = pn) in
        if not (List.exists find_pn !currentPreds) then
          begin
            Util.err "spec" ("undefined pred "^pn);
            Sabsyn.Atom Sabsyn.False
          end
        else
          let f = List.find find_pn !currentPreds in
          let bb = f.Sabsyn.predDefBody in
          let actuals = List.map (de_setExp mn (s @ b)) f2.predprime_form_form in
          let fasubst = List.map2 (fun x y -> (x, y))
              f.Sabsyn.predDefFormals actuals in
          Sabsyn.prime_all (List.fold_right Sabsyn.subst fasubst bb)
    | ANotForm f1 -> Sabsyn.Not (convert_form0 (unsome f1.not_form_form))
    | AAndForm fs -> Sabsyn.And (List.map convert_form0 fs.and_form_form)
    | AOrForm fs -> Sabsyn.Or (List.map convert_form0 fs.or_form_form)
    | AImplForm f1 -> 
        Sabsyn.Impl (convert_form0 (unsome f1.impl_form_lhs),
                     convert_form0 (unsome f1.impl_form_rhs))
    | AIffForm f1 ->
        Sabsyn.Iff (convert_form0 (unsome f1.iff_form_lhs),
                    convert_form0 (unsome f1.iff_form_rhs))
    | AExistsForm f1 ->
        (let f3 = unsome f1.exists_form_form in
        let ABvarDecl f2 = unsome f1.exists_form_bvar_decl in
        let var = s_extract f2.bvar_decl_var in
        let typ = value_type_from_ugly (unsome f2.bvar_decl_type) in
        match typ with
        | TSet s ->
            Sabsyn.ExistsSet ((var,s), convert_form0 f3)
        | _ ->
            Sabsyn.ExistsOne ((var,typ),convert_form0 f3))
    | AForallForm f1 ->
        (let f3 = unsome f1.forall_form_form in
        let ABvarDecl f2 = unsome f1.forall_form_bvar_decl in
        let var = s_extract f2.bvar_decl_var in
        let typ = value_type_from_ugly (unsome f2.bvar_decl_type) in
        match typ with
        | TSet s ->
            Sabsyn.ForallSet ((var,s), convert_form0 f3)
        | _ ->
            Sabsyn.ForallOne ((var,typ),convert_form0 f3))
    | AEqForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ANeqForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ASubForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ACardeqForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ACardleqForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ACardgeqForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ADisjointForm f1 -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | ATrueForm _ -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | AFalseForm _ -> Sabsyn.Atom (de_atom_form mn (s @ b) f)
    | AVarForm s1 -> Sabsyn.Atom (Sabsyn.Propvar 
                                    (de_name mn (s @ b) (unsome s1.var_form_name)))
    | APrimedForm s1 -> 
        (match s1.primed_form_form with
          Some (AVarForm s1) -> Sabsyn.Atom 
              (Sabsyn.Propvarpost (de_name mn (s @ b) (unsome s1.var_form_name)))
        | _ -> failwith "spec deuglifier expecting a post- boolean variable, got funny primed expression")
    | ADiffForm _ -> failwith "spec deuglifier expecting a boolean expression, got set difference"
    | AInterForm _ -> failwith "spec deuglifier expecting a boolean expression, got set intersection"
    | AUnionForm _ -> failwith "spec deuglifier expecting a boolean expression, got set union"
    | AEmptyForm _ -> failwith "spec deuglifier expecting a boolean expression, got emptyset expression" 
    | AUninterpretedStringForm s -> 
        Sabsyn.UninterpretedString (Util.trim_quotes (Tokens.getText 
           (unsome s.uninterpreted_string_form_string_literal))) in
  convert_form0 f
          
(************************************)
(* translating property definitions *)
(************************************)

let de_property mn (spec:p_spec_member) : Sabsyn.predDef list =
  match spec with
  | APropertySpecMember s1 -> 
      let APropertyDecl s2 = (unsome s1.property_spec_member_property_decl) in
      [{
       Sabsyn.predDefName    = s_extract s2.property_decl_name;
       Sabsyn.predDefFormals = List.map Tokens.getText s2.property_decl_args;
       Sabsyn.predDefBody    = convert_form 
         mn (List.map fst !currentSets) []
         (unsome s2.property_decl_body);
      }]
  | _ -> []

let de_properties mn (specs : p_spec_member list)
    : Sabsyn.predDef list = List.concat (List.map (de_property mn) specs)

(*************************************)
(* translating invariant definitions *)
(*************************************)

let de_invariant mn (spec:p_spec_member) : Sabsyn.inv list =
  match spec with
  | AInvSpecMember s1 -> 
      [(convert_form mn
          (List.map fst !currentSets) []
          (unsome s1.inv_spec_member_form),
        if s1.inv_spec_member_public = None then 
          Sabsyn.Private else Sabsyn.Public)]
  | _ -> []

let de_invariants mn (specs : p_spec_member list)
    : Sabsyn.inv list = List.concat (List.map (de_invariant mn) specs)

(************************************)
(* translating pointcut definitions *)
(************************************)

let de_name_star (nn:p_name_star) : Sabsyn.starredIdent = 
  let star_to_string ss = 
    match ss with None -> "" | Some _ -> "*" in
  match nn with
    ASimpleNameStar n1 -> 
      (s_extract n1.simple_name_star_identifier) ^ 
      (star_to_string n1.simple_name_star_star)
  | ASimpleNoNameNameStar n1 -> "*"
  | AQualifiedNameStar n1 ->
      qualToId 
        (s_extract (n1.qualified_name_star_module) ^ 
                    (star_to_string n1.qualified_name_star_mstar))
        (s_extract (n1.qualified_name_star_set) ^
                    (star_to_string n1.qualified_name_star_sstar))
  | AQualifiedNoNameNoModuleNameStar n1 -> "*.*"
  | AQualifiedNoModuleNameStar n1 -> 
      (s_extract (n1.qualified_no_module_name_star_module) ^
       (star_to_string n1.qualified_no_module_name_star_mstar)) ^ ".*"
  | AQualifiedNoNameNameStar n1 -> 
      "*." ^ (s_extract (n1.qualified_no_name_name_star_set) ^
              (star_to_string n1.qualified_no_name_name_star_sstar)) 

let de_proc_arg (pa:p_pc_proc_arg) : Sabsyn.starredType = 
  let APcProcArg p = pa in
  (value_type_from_ugly (unsome p.pc_proc_arg_type), 
   match p.pc_proc_arg_star with
   | None -> false
   | Some _ -> true)

let convert_specpoint_proc (ps:a_pc_proc_spec) : Sabsyn.spProcSpec = 
  let nn = de_name_star (unsome ps.pc_proc_spec_name_star) in
  {
   Sabsyn.sp_module = 
     (let mn = Util.unqualify_getting_module nn in
     if (mn = "") then "this" else mn);
   Sabsyn.sp_name = Util.unqualify nn;
   Sabsyn.sp_args = 
     (match ps.pc_proc_spec_pc_proc_args with 
     | None -> []
     | Some (APcProcArgs a) -> 
         List.map de_proc_arg a.pc_proc_args_pc_proc_arg);
   Sabsyn.sp_retval = 
     (match ps.pc_proc_spec_ret with
     | None -> None
     | Some rv -> Some (de_proc_arg rv));
  }

let rec convert_specpoint w (sp:p_spec_point) : Sabsyn.specPoint = 
  match sp with
  | AMinusSpecPoint p ->
      Sabsyn.SPMinus (convert_specpoint w (unsome p.minus_spec_point_l),
                      convert_specpoint w (unsome p.minus_spec_point_r))
  | AAndSpecPoint p ->
      Sabsyn.SPAnd (convert_specpoint w (unsome p.and_spec_point_l),
                    convert_specpoint w (unsome p.and_spec_point_r))
  | AOrSpecPoint p ->
      Sabsyn.SPOr (convert_specpoint w (unsome p.or_spec_point_l),
                    convert_specpoint w (unsome p.or_spec_point_r))
  | ANotSpecPoint p ->
      Sabsyn.SPNot (convert_specpoint w (unsome p.not_spec_point_spec_point))
  | AProcSpecPoint pp ->
      let APcProcSpec p = unsome pp.proc_spec_point_pc_proc_spec in
      Sabsyn.SPProc (convert_specpoint_proc p)
  | APubModuleSpecPoint p ->
      let n = de_name_star (unsome p.pub_module_spec_point_name_star) in
      if (w = Sabsyn.Scope && n = "this") then 
        Util.err "spec" "can't say module(this) in scope default" else ();
      Sabsyn.SPPubModule n
  | APrivModuleSpecPoint p ->
      let n = de_name_star (unsome p.priv_module_spec_point_name_star) in
      if (w = Sabsyn.Scope && n = "this") then 
        Util.err "spec" "can't say module(this) in scope default" else ();
      Sabsyn.SPPrivModule n
  | AAllModuleSpecPoint p ->
      let n = de_name_star (unsome p.all_module_spec_point_name_star) in
      if (w = Sabsyn.Scope && n = "this") then 
        Util.err "spec" "can't say module(this) in scope default" else ();
      Sabsyn.SPAllModule n
  | APubScopeSpecPoint p ->
      let n = de_name_star (unsome p.pub_scope_spec_point_name_star) in
      if (w = Sabsyn.Module && n = "this") then 
        Util.err "spec" "can't say scope(this) in module default" else ();
      Sabsyn.SPPubScope n
  | APrivScopeSpecPoint p ->
      let n = de_name_star (unsome p.priv_scope_spec_point_name_star) in
      if (w = Sabsyn.Module && n = "this") then 
        Util.err "spec" "can't say scope(this) in module default" else ();
      Sabsyn.SPPrivScope n
  | AAllScopeSpecPoint p ->
      let n = de_name_star (unsome p.all_scope_spec_point_name_star) in
      if (w = Sabsyn.Module && n = "this") then 
        Util.err "spec" "can't say scope(this) in module default" else ();
      Sabsyn.SPAllScope n
  | AAllSpecPoint _ -> Sabsyn.SPAll

let rec convert_pointcut w (pe:p_pointcut_expr) : Sabsyn.pointcutExpr =
  match pe with
  | AMinusPointcutExpr p -> 
      Sabsyn.PCMinus (convert_pointcut w (unsome p.minus_pointcut_expr_l), 
                      convert_pointcut w (unsome p.minus_pointcut_expr_r)) 
  | AAndPointcutExpr p ->
      Sabsyn.PCAnd (convert_pointcut w (unsome p.and_pointcut_expr_l), 
                    convert_pointcut w (unsome p.and_pointcut_expr_r))
  | AOrPointcutExpr p ->
      Sabsyn.PCOr (convert_pointcut w (unsome p.or_pointcut_expr_l), 
                   convert_pointcut w (unsome p.or_pointcut_expr_r)) 
  | ANotPointcutExpr p ->
      Sabsyn.PCNot 
        (convert_pointcut w (unsome p.not_pointcut_expr_pointcut_expr))
  | APrePointcutExpr p ->
      Sabsyn.PCPre 
        (convert_specpoint w (unsome p.pre_pointcut_expr_spec_point))
  | APostPointcutExpr p ->
      Sabsyn.PCPost 
        (convert_specpoint w (unsome p.post_pointcut_expr_spec_point))
  | APrepostPointcutExpr p ->
      Sabsyn.PCPrepost 
        (convert_specpoint w (unsome p.prepost_pointcut_expr_spec_point))
 
(**************************)
(* translating procedures *)
(**************************)

let de_format 
    (spec : p_spec_member) : Sabsyn.typeIdent list =
  match spec with
  | AFormatSpecMember s1 -> 
      (match (unsome s1.format_spec_member_format_decl) with
      AFormatDecl s2 -> 
        (List.map Tokens.getText s2.format_decl_identifier))
  | _ -> []

let rec de_formats (specs : p_spec_member list) 
    : Sabsyn.typeIdent list = List.concat (List.map de_format specs)

let de_specvar
    (spec : p_spec_member)
    : (Id.var_t * Types.value_type) list =
  match spec with
  | ASpecvarSpecMember s1 -> 
      (match (unsome s1.specvar_spec_member_specvar_decl) with
        ASpecvarDecl s2 -> 
          (let names = 
            List.map Tokens.getText s2.specvar_decl_name in
          let typ = value_type_from_ugly (unsome s2.specvar_decl_type) in
          let addType n = (n,typ) in
          (List.map addType names)))
  | _ -> []

let rec de_specvars
    (specs : p_spec_member list) 
    : (Id.var_t * Types.value_type) list =
   currentSets := List.concat (List.map de_specvar specs);
   !currentSets

let de_formal (f : p_formal) 
    : (Id.var_t * Types.value_type) =
  match f with AFormal f1 -> 
    (Id.fetch_var (s_extract f1.formal_fname),
     value_type_from_ugly (unsome f1.formal_type))

let de_retVal (r : p_returns_decl option) 
    : (Id.var_t * Types.value_type) option =
  match r with
  | None -> None
  | Some (AReturnsDecl rv) -> 
      Some (Id.fetch_var (s_extract rv.returns_decl_rname),
            value_type_from_ugly (unsome rv.returns_decl_type))

let de_suspends mn (s : p_suspends_decl option) : Id.default_t list =
  match s with
  | None -> []
  | Some (ASuspendsDecl { suspends_decl_name = ss }) -> 
      List.map (fun s0 ->
        match s0 with
        | ASimpleName n1 -> 
            Id.fetch_default mn (s_extract n1.simple_name_identifier)
        | AQualifiedName n1 ->
            Id.fetch_default (s_extract n1.qualified_name_module)
              (s_extract n1.qualified_name_set)) ss
    
let de_requires mn (req : p_requires_decl option) : Sabsyn.form =
  match req with
  | None -> Sabsyn.Atom (Sabsyn.True)
  | Some (ARequiresDecl f1) ->
      convert_form 
        mn (List.map fst !currentSets) []
        (unsome f1.requires_decl_form)

let de_modifies mn qn (modi : (p_modifies_decl option)) 
    : Id.module_t list =
  match modi with
  | None -> []
  | Some (AModifiesDecl { modifies_decl_name = tks }) -> 
      List.map (de_name mn qn) tks

let de_calls (cal : p_calls_decl option)
    : Id.module_t list =
  match cal with
  | None -> []
  | Some (ACallsDecl {calls_decl_callee_decl = cls}) ->
      let getCallee x =
        match x with 
          (ACalleeDecl {callee_decl_identifier = Some tok}) 
          -> Tokens.getText tok 
        | _ -> internal "Unexpected abstract syntax of mod clause"
      in 
      List.map getCallee cls

let de_ensures mn (ens : p_ensures_decl option) : Sabsyn.form =
  match ens with
  | None -> Sabsyn.Atom (Sabsyn.True)
  | Some (AEnsuresDecl ed) -> 
      convert_form 
        mn (List.map fst !currentSets) []
        (unsome ed.ensures_decl_form)

let de_default w mn (d : p_default_decl) : Sabsyn.defaultDef =
  let ADefaultDecl s2 = d in
  {
   Sabsyn.defaultName     = Id.fetch_default mn
                                (s_extract s2.default_decl_name);
   Sabsyn.defaultLocation = w;
   Sabsyn.defaultParms    = List.map (de_name mn []) s2.default_decl_args;
   Sabsyn.defaultPointcut = 
   (match s2.default_decl_pc with
   | None -> Sabsyn.PCPre 
         (if (w = Sabsyn.Module) then 
           (Sabsyn.SPAllModule "this") else (Sabsyn.SPAllScope "this"))
   | Some dd -> convert_pointcut w dd);
   Sabsyn.defaultBody     = convert_form 
     mn (List.map fst !currentSets) []
     (unsome s2.default_decl_body);
 }

let de_defaults w mn (spec : p_spec_member list) : Sabsyn.defaultDef list =
  List.concat (List.map (fun s -> match s with
  | ADefaultSpecMember s1 -> 
      [de_default w mn (unsome s1.default_spec_member_default_decl)]
  | _ -> []) spec)

let de_proc mn qn
    (spec : p_spec_member) : Sabsyn.proc_spec list =
  match spec with
  | AProcSpecMember s1 ->
      (match (unsome s1.proc_spec_member_proc_decl) with
        AProcDecl s2 ->
          [{ Sabsyn.proc_name = 
             { Id.p_name = s_extract s2.proc_decl_identifier;
               Id.p_module = mn };
             Sabsyn.proc_modifiers = {
             Id.is_private = match (s2.proc_decl_private) with
             | None -> false | Some _ -> true; };
             Sabsyn.formals = List.map de_formal s2.proc_decl_formal;
             Sabsyn.ret_val = de_retVal s2.proc_decl_returns_decl;
             Sabsyn.suspends = de_suspends mn s2.proc_decl_suspends_decl;
             Sabsyn.requires = de_requires mn s2.proc_decl_requires_decl;
             Sabsyn.modifies = de_modifies mn qn s2.proc_decl_modifies_decl;
             Sabsyn.calls = de_calls s2.proc_decl_calls_decl;
             Sabsyn.extra_ensures = [];
             Sabsyn.ensures = de_ensures mn s2.proc_decl_ensures_decl
           }])
  | _ -> []

let rec de_procs mn qn (specs : p_spec_member list) 
    : Sabsyn.proc_spec list =
   List.concat (List.map (de_proc mn qn) specs)

(****************************)
(* translating spec modules *)
(****************************)

let convert (m : p_spec_module) : Sabsyn.spec_module =  
   match m with 
   | AModSpecModule m1 ->
       let name = s_extract m1.mod_spec_module_identifier in
       currentSets := [];
       currentPreds := [];
       let fmts  = de_formats (m1.mod_spec_module_spec_member) in
       let specvars = de_specvars (m1.mod_spec_module_spec_member) in
       let defaults = de_defaults 
           Sabsyn.Module name (m1.mod_spec_module_spec_member) in
       let props = de_properties name (m1.mod_spec_module_spec_member) in
       let invariants = de_invariants name (m1.mod_spec_module_spec_member) in
       currentPreds := props;
       let procs = de_procs name (List.map fst specvars) 
           (m1.mod_spec_module_spec_member) in
       { 
         Sabsyn.module_name = Id.fetch_module name;
         Sabsyn.instantiated_from = None; 
         Sabsyn.param_subst = [];
         Sabsyn.formats = fmts;
         Sabsyn.properties = props;
         Sabsyn.invariants = invariants;
         Sabsyn.specvars = specvars;
         Sabsyn.procs = procs;
         Sabsyn.defaults = defaults;
       }
   | AInstSpecModule m1 ->  
       { 
         Sabsyn.module_name = Id.fetch_module 
           (s_extract m1.inst_spec_module_n);
         Sabsyn.instantiated_from = Some 
           (Id.fetch_module (s_extract m1.inst_spec_module_parent));
         Sabsyn.param_subst = List.map (fun (ASubst x) -> 
          (Types.TObj (Id.fetch_format (s_extract x.subst_f)),
           Types.TObj (Id.fetch_format (s_extract x.subst_a))))
          m1.inst_spec_module_params;
         Sabsyn.formats = [];
         Sabsyn.properties = [];
         Sabsyn.invariants = [];
         Sabsyn.specvars = [];
         Sabsyn.procs = [];
         Sabsyn.defaults = [];
       }

let convert_scope (m0 : p_scope_decl) : Sabsyn.scope =  
   match m0 with 
   | AScopeDecl m ->
       let name = s_extract m.scope_decl_name in
       currentSets := [];
       { 
         Sabsyn.scope_name = Id.fetch_scope name;
         Sabsyn.modules    = (match m.scope_decl_modules_decl with
         | None -> []
         | Some (AModulesDecl a) -> 
             (List.map Tokens.getText a.modules_decl_identifier));
         Sabsyn.exports    = (match m.scope_decl_exports_decl with
         | None -> []
         | Some (AExportsDecl a) -> 
             (List.map Tokens.getText a.exports_decl_identifier));
         Sabsyn.invariant  = (match m.scope_decl_spec_member with
         | None -> None | Some (AInvSpecMember f) -> 
             Some (convert_form name [] [] 
                     (unsome (f.inv_spec_member_form)),
                   if f.inv_spec_member_public = None then 
                     Sabsyn.Private else Sabsyn.Public) | _ -> None);
         Sabsyn.scope_defaults = 
             (List.map (de_default Sabsyn.Scope name) 
                m.scope_decl_default_decl);
       }

